TIMERS_SIZE = 212
CMD_BUFFER_CORE_SIZE = 264
DYNA_SIZE = 1424
SCENE_MOBIL_SIZE = 2168
SIMULATION_WHEELS_SIZE = 3056
PLUG_SOLID_SIZE = 68
PLAYER_INFO_SIZE = 952
INPUT_STATE_SIZE = 120

SIM_HAS_TIMERS = 0x1
SIM_HAS_DYNA = 0x2
SIM_HAS_SCENE_MOBIL = 0x4
SIM_HAS_SIMULATION_WHEELS = 0x8
SIM_HAS_PLUG_SOLID = 0x10
SIM_HAS_CMD_BUFFER_CORE = 0x20
SIM_HAS_INPUT_STATE = 0x40
SIM_HAS_PLAYER_INFO = 0x80

MODE_SIMULATION = 0
MODE_RUN = 1

BINARY_ACCELERATE_NAME = 'Accelerate'
BINARY_BRAKE_NAME = 'Brake'
BINARY_LEFT_NAME = 'Steer left'
BINARY_RIGHT_NAME = 'Steer right'
BINARY_RESPAWN_NAME = 'Respawn'
BINARY_HORN_NAME = 'Horn'
BINARY_RACE_START_NAME = '_FakeIsRaceRunning'
BINARY_RACE_FINISH_NAME = '_FakeFinishLine'
ANALOG_STEER_NAME = 'Steer (analog)'
ANALOG_ACCELERATE_NAME = 'Accelerate (analog)'

DEFAULT_SERVER_SIZE = 65535